__version__ = "2.9.2"
# from .coretypes import BoxPadMethod, Coordinate3D, RegionBox, Size3D, Spacing3D
# from .datasets import example_data, example_data_paths
# from .dicom import find_dicoms, lookup_tag, similar_tags, tag_exists
# from .loggers import logger

# __all__ = [
#     "find_dicoms",
#     "lookup_tag",
#     "similar_tags",
#     "tag_exists",
#     "logger",
#     ## coretypes
#     "BoxPadMethod",
#     "RegionBox",
#     "Coordinate3D",
#     "Size3D",
#     "Spacing3D",
#     ## example data
#     "example_data",
#     "example_data_paths",
# ]
